@@ -1,5 +1,5 @@
-// LNP/BP client-side-validation library implementing respective LNPBP
-// specifications & standards (LNPBP-7, 8, 9, 42)
+// LNP/BP client-side-validation foundation libraries implementing LNPBP
+// specifications & standards (LNPBP-4, 7, 8, 9, 42, 81)
 //
 // Written in 2019-2021 by
 //     Dr. Maxim Orlovsky <orlovsky@pandoracore.com>
@@ -15,7 +15,6 @@
 use std::io;
 
 use bitcoin::bech32::u5;
-use bitcoin::hashes::{hash160, hmac, sha256, sha256d, sha256t, sha512, Hash};
 use bitcoin::util::address::{self, Address};
 use bitcoin::util::psbt::PartiallySignedTransaction;
 use bitcoin::{
@@ -54,73 +53,6 @@ impl Strategy for SigHash {
     type Strategy = strategies::HashFixedBytes;
 }
 
-impl Strategy for sha256::Hash {
-    type Strategy = strategies::HashFixedBytes;
-}
-impl Strategy for sha256d::Hash {
-    type Strategy = strategies::HashFixedBytes;
-}
-impl<T> Strategy for sha256t::Hash<T>
-where
-    T: sha256t::Tag,
-{
-    type Strategy = strategies::HashFixedBytes;
-}
-impl Strategy for sha512::Hash {
-    type Strategy = strategies::HashFixedBytes;
-}
-impl Strategy for hash160::Hash {
-    type Strategy = strategies::HashFixedBytes;
-}
-impl<T> Strategy for hmac::Hmac<T>
-where
-    T: Hash,
-{
-    type Strategy = strategies::HashFixedBytes;
-}
-
-impl Strategy for OutPoint {
-    type Strategy = strategies::BitcoinConsensus;
-}
-impl Strategy for TxOut {
-    type Strategy = strategies::BitcoinConsensus;
-}
-impl Strategy for TxIn {
-    type Strategy = strategies::BitcoinConsensus;
-}
-impl Strategy for Transaction {
-    type Strategy = strategies::BitcoinConsensus;
-}
-impl Strategy for PartiallySignedTransaction {
-    type Strategy = strategies::BitcoinConsensus;
-}
-
-impl StrictEncode for Amount {
-    fn strict_encode<E: io::Write>(&self, e: E) -> Result<usize, Error> {
-        self.as_sat().strict_encode(e)
-    }
-}
-
-impl StrictDecode for Amount {
-    fn strict_decode<D: io::Read>(d: D) -> Result<Self, Error> {
-        Ok(Amount::from_sat(u64::strict_decode(d)?))
-    }
-}
-
-impl StrictEncode for Script {
-    #[inline]
-    fn strict_encode<E: io::Write>(&self, e: E) -> Result<usize, Error> {
-        self.to_bytes().strict_encode(e)
-    }
-}
-
-impl StrictDecode for Script {
-    #[inline]
-    fn strict_decode<D: io::Read>(d: D) -> Result<Self, Error> {
-        Ok(Self::from(Vec::<u8>::strict_decode(d)?))
-    }
-}
-
 impl StrictEncode for secp256k1::SecretKey {
     #[inline]
     fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
@@ -151,8 +83,14 @@ impl StrictDecode for secp256k1::PublicKey {
     fn strict_decode<D: io::Read>(mut d: D) -> Result<Self, Error> {
         let mut buf = [0u8; secp256k1::constants::PUBLIC_KEY_SIZE];
         d.read_exact(&mut buf)?;
+        if buf[0] == 0x04 {
+            return Err(Error::DataIntegrityError(s!(
+                "invalid public key data: uncompressed Secp256k1 public key \
+                format is not allowed, use compressed form instead"
+            )));
+        }
         Self::from_slice(&buf).map_err(|_| {
-            Error::DataIntegrityError("invalid public key data".to_string())
+            Error::DataIntegrityError(s!("invalid public key data"))
         })
     }
 }
@@ -160,21 +98,32 @@ impl StrictDecode for secp256k1::PublicKey {
 impl StrictEncode for secp256k1::schnorrsig::PublicKey {
     #[inline]
     fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
-        Ok(e.write(&self.serialize())?)
+        Ok(e.write(&[0x02_u8])? + e.write(&self.serialize())?)
     }
 }
 
 impl StrictDecode for secp256k1::schnorrsig::PublicKey {
     #[inline]
     fn strict_decode<D: io::Read>(mut d: D) -> Result<Self, Error> {
-        let mut buf = [0u8; secp256k1::constants::SCHNORRSIG_PUBLIC_KEY_SIZE];
+        let mut buf =
+            [0u8; secp256k1::constants::SCHNORRSIG_PUBLIC_KEY_SIZE + 1];
         d.read_exact(&mut buf)?;
-        Self::from_slice(&buf).map_err(|_| {
-            Error::DataIntegrityError("invalid public key data".to_string())
+        if buf[0] != 0x02 {
+            return Err(Error::DataIntegrityError(s!(
+                "invalid public key data: BIP340 keys must be serialized \
+                with `0x02` prefix byte"
+            )));
+        }
+        Self::from_slice(&buf[1..]).map_err(|_| {
+            Error::DataIntegrityError(s!("invalid public key data"))
         })
     }
 }
 
+// TODO: #17 Implement strict encoding for `KeyPair` type once there will be a
+//       way to serialize its inner data in Secpk256k1 lib (see
+//       <https://github.com/rust-bitcoin/rust-secp256k1/issues/298>)
+
 impl StrictEncode for secp256k1::Signature {
     #[inline]
     fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
@@ -189,7 +138,7 @@ impl StrictDecode for secp256k1::Signature {
         d.read_exact(&mut buf)?;
         Self::from_compact(&buf).map_err(|_| {
             Error::DataIntegrityError(
-                "Invalid secp256k1 signature data".to_string(),
+                "Invalid secp256k1 ECDSA signature data".to_string(),
             )
         })
     }
@@ -209,12 +158,19 @@ impl StrictDecode for secp256k1::schnorrsig::Signature {
         d.read_exact(&mut buf)?;
         Self::from_slice(&buf).map_err(|_| {
             Error::DataIntegrityError(
-                "Invalid secp256k1 signature data".to_string(),
+                "Invalid secp256k1 Schnorr signature data".to_string(),
             )
         })
     }
 }
 
+#[doc(hidden)]
+#[allow(useless_deprecated)]
+#[deprecated(
+    since = "1.4.0",
+    note = "Uncompressed PublicKey serialization is not recommended, use \
+            `secp256k1::PublicKey` type instead"
+)]
 impl StrictEncode for bitcoin::PublicKey {
     #[inline]
     fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
@@ -226,6 +182,13 @@ impl StrictEncode for bitcoin::PublicKey {
     }
 }
 
+#[doc(hidden)]
+#[allow(useless_deprecated)]
+#[deprecated(
+    since = "1.4.0",
+    note = "Uncompressed PublicKey serialization is not recommended, use \
+            `secp256k1::PublicKey` type instead"
+)]
 impl StrictDecode for bitcoin::PublicKey {
     #[inline]
     fn strict_decode<D: io::Read>(mut d: D) -> Result<Self, Error> {
@@ -248,13 +211,110 @@ impl StrictDecode for bitcoin::PublicKey {
                 })?)
             }
             invalid_flag => Err(Error::DataIntegrityError(format!(
-                "Invalid public key encoding flag {}; must be either 0x02, 0x03 or 0x4",
+                "Invalid public key encoding flag {:#04x}; must be either 0x02, 0x03 or 0x04",
                 invalid_flag
             ))),
         }
     }
 }
 
+impl Strategy for OutPoint {
+    type Strategy = strategies::BitcoinConsensus;
+}
+impl Strategy for TxOut {
+    type Strategy = strategies::BitcoinConsensus;
+}
+impl Strategy for TxIn {
+    type Strategy = strategies::BitcoinConsensus;
+}
+impl Strategy for Transaction {
+    type Strategy = strategies::BitcoinConsensus;
+}
+impl Strategy for PartiallySignedTransaction {
+    type Strategy = strategies::BitcoinConsensus;
+}
+
+impl StrictEncode for address::Payload {
+    fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
+        Ok(match self {
+            address::Payload::PubkeyHash(pkh) => {
+                32u8.strict_encode(&mut e)? + pkh.strict_encode(&mut e)?
+            }
+            address::Payload::ScriptHash(sh) => {
+                33u8.strict_encode(&mut e)? + sh.strict_encode(&mut e)?
+            }
+            address::Payload::WitnessProgram { version, program } => {
+                version.to_u8().strict_encode(&mut e)?
+                    + program.strict_encode(&mut e)?
+            }
+        })
+    }
+}
+
+impl StrictDecode for address::Payload {
+    fn strict_decode<D: io::Read>(mut d: D) -> Result<Self, Error> {
+        Ok(match u8::strict_decode(&mut d)? {
+            32u8 => {
+                address::Payload::PubkeyHash(PubkeyHash::strict_decode(&mut d)?)
+            }
+            33u8 => {
+                address::Payload::ScriptHash(ScriptHash::strict_decode(&mut d)?)
+            }
+            // TODO: #18 Update to `WitnessVersion` upon bitcoin 0.26.1 release
+            version if version <= 16 => address::Payload::WitnessProgram {
+                version: u5::try_from_u8(version)
+                    .expect("bech32::u8 decider is broken"),
+                program: StrictDecode::strict_decode(&mut d)?,
+            },
+            wrong => {
+                return Err(Error::ValueOutOfRange(
+                    "witness program version",
+                    0..17,
+                    wrong as u128,
+                ))
+            }
+        })
+    }
+}
+
+impl StrictEncode for Address {
+    fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
+        Ok(strict_encode_list!(e; self.network, self.payload))
+    }
+}
+
+impl StrictDecode for Address {
+    fn strict_decode<D: io::Read>(mut d: D) -> Result<Self, Error> {
+        Ok(strict_decode_self!(d; network, payload; crate))
+    }
+}
+
+impl StrictEncode for Amount {
+    fn strict_encode<E: io::Write>(&self, e: E) -> Result<usize, Error> {
+        self.as_sat().strict_encode(e)
+    }
+}
+
+impl StrictDecode for Amount {
+    fn strict_decode<D: io::Read>(d: D) -> Result<Self, Error> {
+        Ok(Amount::from_sat(u64::strict_decode(d)?))
+    }
+}
+
+impl StrictEncode for Script {
+    #[inline]
+    fn strict_encode<E: io::Write>(&self, e: E) -> Result<usize, Error> {
+        self.to_bytes().strict_encode(e)
+    }
+}
+
+impl StrictDecode for Script {
+    #[inline]
+    fn strict_decode<D: io::Read>(d: D) -> Result<Self, Error> {
+        Ok(Self::from(Vec::<u8>::strict_decode(d)?))
+    }
+}
+
 impl StrictEncode for bitcoin::Network {
     #[inline]
     fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
@@ -388,84 +448,140 @@ impl StrictDecode for bip32::ExtendedPrivKey {
     }
 }
 
-impl StrictEncode for Address {
-    fn strict_encode<E: io::Write>(&self, mut e: E) -> Result<usize, Error> {
-        let mut len = 0usize;
-        len += self.network.strict_encode(&mut e)?;
-        match &self.payload {
-            address::Payload::PubkeyHash(pkh) => {
-                len += 32u8.strict_encode(&mut e)?;
-                len += pkh.strict_encode(&mut e)?;
-            }
-            address::Payload::ScriptHash(sh) => {
-                len += 33u8.strict_encode(&mut e)?;
-                len += sh.strict_encode(&mut e)?;
-            }
-            address::Payload::WitnessProgram { version, program } => {
-                len += version.to_u8().strict_encode(&mut e)?;
-                len += program.strict_encode(&mut e)?;
-            }
-        };
-        Ok(len)
-    }
-}
-
-impl StrictDecode for Address {
-    fn strict_decode<D: io::Read>(mut d: D) -> Result<Self, Error> {
-        let network = bitcoin::Network::strict_decode(&mut d)?;
-        let payload = match u8::strict_decode(&mut d)? {
-            32u8 => {
-                address::Payload::PubkeyHash(PubkeyHash::strict_decode(&mut d)?)
-            }
-            33u8 => {
-                address::Payload::ScriptHash(ScriptHash::strict_decode(&mut d)?)
-            }
-            version => address::Payload::WitnessProgram {
-                version: u5::try_from_u8(version).map_err(|_| {
-                    Error::ValueOutOfRange(
-                        "witness program version",
-                        0..17,
-                        version as u128,
-                    )
-                })?,
-                program: StrictDecode::strict_decode(&mut d)?,
-            },
-        };
-        Ok(Address { payload, network })
-    }
-}
-
 #[cfg(test)]
 pub(crate) mod test {
     use std::str::FromStr;
 
-    use bitcoin::{consensus, hashes::hex::FromHex, secp256k1::Message};
+    use bitcoin::{
+        consensus, hashes::hex::FromHex, hashes::Hash, secp256k1::Message,
+    };
 
     use super::*;
-    use crate::strict_serialize;
-    use crate::test_helpers::test_suite;
+    use crate::test_helpers::*;
 
     #[test]
-    fn test_encoding_network() {
-        let mainnet_bytes = &[0xF9u8, 0xBEu8, 0xB4u8, 0xD9u8][..];
-        let testnet_bytes = &[0x0Bu8, 0x11u8, 0x09u8, 0x07u8][..];
-        let regtest_bytes = &[0xFAu8, 0xBFu8, 0xB5u8, 0xDAu8][..];
+    fn test_encoding_hashes() {
+        static HASH256_BYTES: [u8; 32] = [
+            0x15, 0x2d, 0x1c, 0x97, 0x61, 0xd4, 0x64, 0x66, 0x68, 0xdf, 0xcd,
+            0xeb, 0x11, 0x98, 0x70, 0x84, 0x4e, 0xdb, 0x25, 0xa0, 0xea, 0x1e,
+            0x35, 0x20, 0x7f, 0xaa, 0x44, 0xa9, 0x67, 0xa6, 0xa6, 0x61,
+        ];
+        static HASH160_BYTES: [u8; 20] = [
+            0x15, 0x2d, 0x1c, 0x97, 0x61, 0xd4, 0x64, 0x66, 0x68, 0xdf, 0xcd,
+            0xeb, 0x11, 0x98, 0x4e, 0xdb, 0x25, 0xa0, 0xea, 0x1e,
+        ];
 
-        let mainnet = bitcoin::Network::strict_decode(mainnet_bytes).unwrap();
-        let testnet = bitcoin::Network::strict_decode(testnet_bytes).unwrap();
-        let regtest = bitcoin::Network::strict_decode(regtest_bytes).unwrap();
+        const TEST_MIDSTATE: [u8; 32] = [
+            156, 224, 228, 230, 124, 17, 108, 57, 56, 179, 202, 242, 195, 15,
+            80, 137, 211, 243, 147, 108, 71, 99, 110, 96, 125, 179, 62, 234,
+            221, 198, 240, 201,
+        ];
 
-        test_suite(&mainnet, &mainnet_bytes, 4);
-        test_suite(&testnet, &testnet_bytes, 4);
-        test_suite(&regtest, &regtest_bytes, 4);
+        #[derive(
+            Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Default, Hash,
+        )]
+        pub struct TestHashTag;
+
+        impl sha256t::Tag for TestHashTag {
+            fn engine() -> sha256::HashEngine {
+                // The TapRoot TapLeaf midstate.
+                let midstate = sha256::Midstate::from_inner(TEST_MIDSTATE);
+                sha256::HashEngine::from_midstate(midstate, 64)
+            }
+        }
+
+        test_encoding_roundtrip(
+            &ripemd160::Hash::from_inner(HASH160_BYTES),
+            HASH160_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &hash160::Hash::from_inner(HASH160_BYTES),
+            HASH160_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &hmac::Hmac::<sha256::Hash>::from_inner(HASH256_BYTES),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &sha256::Hash::from_inner(HASH256_BYTES),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &sha256d::Hash::from_inner(HASH256_BYTES),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &sha256t::Hash::<TestHashTag>::from_inner(HASH256_BYTES),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &Txid::from_slice(&HASH256_BYTES).unwrap(),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &Wtxid::from_slice(&HASH256_BYTES).unwrap(),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &BlockHash::from_slice(&HASH256_BYTES).unwrap(),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &XpubIdentifier::from_slice(&HASH160_BYTES).unwrap(),
+            HASH160_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &SigHash::from_slice(&HASH256_BYTES).unwrap(),
+            HASH256_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &PubkeyHash::from_slice(&HASH160_BYTES).unwrap(),
+            HASH160_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &WPubkeyHash::from_slice(&HASH160_BYTES).unwrap(),
+            HASH160_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &ScriptHash::from_slice(&HASH160_BYTES).unwrap(),
+            HASH160_BYTES,
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &WScriptHash::from_slice(&HASH256_BYTES).unwrap(),
+            HASH256_BYTES,
+        )
+        .unwrap();
     }
 
     #[test]
-    #[should_panic]
-    fn test_encoding_network_failure() {
-        // Bitcoin Network structure do not support "Other" networks
-        let random_bytes = &[0xA1u8, 0xA2u8, 0xA3u8, 0xA4u8][..];
-        bitcoin::Network::strict_decode(random_bytes).unwrap();
+    fn test_encoding_seckey(
+    ) -> Result<(), DataEncodingTestFailure<secp256k1::SecretKey>> {
+        let secp = secp256k1::Secp256k1::new();
+        static SK_BYTES: [u8; 32] = [
+            0x15, 0x2d, 0x1c, 0x97, 0x61, 0xd4, 0x64, 0x66, 0x68, 0xdf, 0xcd,
+            0xeb, 0x11, 0x98, 0x70, 0x84, 0x4e, 0xdb, 0x25, 0xa0, 0xea, 0x1e,
+            0x35, 0x20, 0x7f, 0xaa, 0x44, 0xa9, 0x67, 0xa6, 0xa6, 0x61,
+        ];
+        let sk = secp256k1::SecretKey::from_slice(&SK_BYTES).unwrap();
+        let _sk_bip340 =
+            secp256k1::schnorrsig::KeyPair::from_seckey_slice(&secp, &SK_BYTES)
+                .unwrap();
+        // TODO: #17 implement KeyPair serialization testing
+        test_encoding_roundtrip(&sk, &SK_BYTES[..])
     }
 
     #[test]
@@ -489,24 +605,76 @@ pub(crate) mod test {
             0x9e, 0x4e, 0xe9, 0x71, 0xd8, 0x6b, 0x5e, 0x61, 0x87, 0x5d,
         ];
         static PK_BYTES_ONEKEY: [u8; 33] = [
-            0x2, 0x79, 0xbe, 0x66, 0x7e, 0xf9, 0xdc, 0xbb, 0xac, 0x55, 0xa0,
-            0x62, 0x95, 0xce, 0x87, 0xb, 0x7, 0x2, 0x9b, 0xfc, 0xdb, 0x2d,
+            0x02, 0x79, 0xbe, 0x66, 0x7e, 0xf9, 0xdc, 0xbb, 0xac, 0x55, 0xa0,
+            0x62, 0x95, 0xce, 0x87, 0x0b, 0x07, 0x02, 0x9b, 0xfc, 0xdb, 0x2d,
             0xce, 0x28, 0xd9, 0x59, 0xf2, 0x81, 0x5b, 0x16, 0xf8, 0x17, 0x98,
         ];
 
-        let pubkey_02 =
-            bitcoin::PublicKey::strict_decode(&PK_BYTES_02[..]).unwrap();
-        let pubkey_03 =
-            bitcoin::PublicKey::strict_decode(&PK_BYTES_03[..]).unwrap();
-        let pubkey_04 =
-            bitcoin::PublicKey::strict_decode(&PK_BYTES_04[..]).unwrap();
-        let pubkey_onekey =
-            bitcoin::PublicKey::strict_decode(&PK_BYTES_ONEKEY[..]).unwrap();
+        let secp_pk_02 =
+            secp256k1::PublicKey::from_slice(&PK_BYTES_02).unwrap();
+        let secp_pk_03 =
+            secp256k1::PublicKey::from_slice(&PK_BYTES_03).unwrap();
+        let secp_pk_one =
+            secp256k1::PublicKey::from_slice(&PK_BYTES_ONEKEY).unwrap();
+        test_encoding_roundtrip(&secp_pk_02, PK_BYTES_02).unwrap();
+        test_encoding_roundtrip(&secp_pk_03, PK_BYTES_03).unwrap();
+        test_encoding_roundtrip(&secp_pk_one, PK_BYTES_ONEKEY).unwrap();
+        assert_eq!(
+            secp256k1::PublicKey::strict_deserialize(&PK_BYTES_04),
+            Err(Error::DataIntegrityError(s!(
+                "invalid public key data: uncompressed Secp256k1 public key \
+                format is not allowed, use compressed form instead"
+            )))
+        );
 
-        test_suite(&pubkey_02, &PK_BYTES_02, 33);
-        test_suite(&pubkey_03, &PK_BYTES_03, 33);
-        test_suite(&pubkey_04, &PK_BYTES_04, 65);
-        test_suite(&pubkey_onekey, &PK_BYTES_ONEKEY, 33);
+        let sk_one = secp256k1::PublicKey::from_secret_key(
+            &secp256k1::Secp256k1::new(),
+            &secp256k1::key::ONE_KEY,
+        );
+
+        let pubkey_02 = bitcoin::PublicKey::from_slice(&PK_BYTES_02).unwrap();
+        let pubkey_03 = bitcoin::PublicKey::from_slice(&PK_BYTES_03).unwrap();
+        let pubkey_04 = bitcoin::PublicKey::from_slice(&PK_BYTES_04).unwrap();
+        let one_key = bitcoin::PublicKey {
+            compressed: true,
+            key: sk_one,
+        };
+        test_encoding_roundtrip(&pubkey_02, PK_BYTES_02).unwrap();
+        test_encoding_roundtrip(&pubkey_03, PK_BYTES_03).unwrap();
+        test_encoding_roundtrip(&pubkey_04, PK_BYTES_04).unwrap();
+        test_encoding_roundtrip(&one_key, PK_BYTES_ONEKEY).unwrap();
+        assert_eq!(secp_pk_02, pubkey_02.key);
+        assert_eq!(secp_pk_02, pubkey_02.key);
+        assert_eq!(secp_pk_02, pubkey_02.key);
+        assert_eq!(secp_pk_03, pubkey_03.key);
+        assert_eq!(secp_pk_one, one_key.key);
+        assert_eq!(pubkey_03.key, pubkey_04.key);
+
+        let xcoordonly_02 =
+            secp256k1::schnorrsig::PublicKey::from_slice(&PK_BYTES_02[1..])
+                .unwrap();
+        let xcoordonly_one =
+            secp256k1::schnorrsig::PublicKey::from_slice(&PK_BYTES_ONEKEY[1..])
+                .unwrap();
+        test_encoding_roundtrip(&xcoordonly_02, PK_BYTES_02).unwrap();
+        test_encoding_roundtrip(&xcoordonly_one, PK_BYTES_ONEKEY).unwrap();
+        assert_eq!(
+            secp256k1::schnorrsig::PublicKey::strict_decode(&PK_BYTES_03[..]),
+            Err(Error::DataIntegrityError(s!(
+                "invalid public key data: BIP340 keys must be serialized \
+                with `0x02` prefix byte"
+            )))
+        );
+        assert_eq!(
+            secp256k1::schnorrsig::PublicKey::strict_decode(&PK_BYTES_04[..]),
+            Err(Error::DataIntegrityError(s!(
+                "invalid public key data: BIP340 keys must be serialized \
+                with `0x02` prefix byte"
+            )))
+        );
+        assert_eq!(xcoordonly_02.serialize(), secp_pk_02.serialize()[1..]);
+        assert_eq!(xcoordonly_02.serialize(), secp_pk_03.serialize()[1..]);
+        assert_eq!(xcoordonly_one.serialize(), secp_pk_one.serialize()[1..]);
     }
 
     #[test]
@@ -534,47 +702,223 @@ pub(crate) mod test {
         bitcoin::PublicKey::strict_decode(&PK_BYTES_02[..]).unwrap();
     }
 
+    static ECDSA_BYTES: [u8; 64] = [
+        0xdf, 0x2b, 0x07, 0x01, 0x5f, 0x2e, 0x01, 0x67, 0x74, 0x18, 0x7e, 0xad,
+        0x4a, 0x4f, 0x71, 0x9a, 0x14, 0xe3, 0xe1, 0xad, 0xa1, 0x78, 0xd6, 0x6c,
+        0xce, 0xcf, 0xa4, 0x5b, 0x63, 0x30, 0x70, 0xc2, 0x43, 0xa2, 0xd7, 0x6e,
+        0xe0, 0x5d, 0x63, 0x49, 0xfe, 0x98, 0x69, 0x6c, 0x1c, 0x4d, 0x9a, 0x67,
+        0x11, 0x24, 0xde, 0x40, 0xc5, 0x31, 0x71, 0xa4, 0xb2, 0x82, 0xb7, 0x69,
+        0xb7, 0xc6, 0x96, 0xcd,
+    ];
+
+    static SCHNORR_BYTES: [u8; 64] = [
+        0x16, 0x2A, 0xDD, 0x99, 0x90, 0x6A, 0x93, 0x04, 0xD4, 0xD0, 0x79, 0x02,
+        0x73, 0xAA, 0xD1, 0xB3, 0x54, 0x13, 0x6C, 0x32, 0xBE, 0x53, 0xF0, 0x7D,
+        0xF8, 0x0F, 0x74, 0xB3, 0xF7, 0xBF, 0x99, 0xAE, 0x91, 0x13, 0x45, 0x76,
+        0xD0, 0x4B, 0x8A, 0x2D, 0x82, 0xDA, 0x7F, 0xE9, 0xDA, 0x43, 0x99, 0xF3,
+        0xBD, 0x99, 0xDF, 0x6B, 0xA3, 0xC6, 0xF9, 0xB4, 0xFB, 0x5E, 0x62, 0xDB,
+        0x1B, 0x15, 0xCF, 0xF3,
+    ];
+
     #[test]
-    fn test_encode_signature() {
-        let s = secp256k1::Secp256k1::new();
+    fn test_encode_signatures() {
+        let secp = secp256k1::Secp256k1::new();
 
-        static SIG_BYTES: [u8; 64] = [
-            0xdf, 0x2b, 0x7, 0x1, 0x5f, 0x2e, 0x1, 0x67, 0x74, 0x18, 0x7e,
-            0xad, 0x4a, 0x4f, 0x71, 0x9a, 0x14, 0xe3, 0xe1, 0xad, 0xa1, 0x78,
-            0xd6, 0x6c, 0xce, 0xcf, 0xa4, 0x5b, 0x63, 0x30, 0x70, 0xc2, 0x43,
-            0xa2, 0xd7, 0x6e, 0xe0, 0x5d, 0x63, 0x49, 0xfe, 0x98, 0x69, 0x6c,
-            0x1c, 0x4d, 0x9a, 0x67, 0x11, 0x24, 0xde, 0x40, 0xc5, 0x31, 0x71,
-            0xa4, 0xb2, 0x82, 0xb7, 0x69, 0xb7, 0xc6, 0x96, 0xcd,
-        ];
-
-        let privkey = secp256k1::SecretKey::from_slice(&[
+        static KEY: [u8; 32] = [
             0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
             0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xBA, 0xAE, 0xDC, 0xE6, 0xAF, 0x48,
             0xA0, 0x3B, 0xBF, 0xD2, 0x5E, 0x8C, 0xD0, 0x36, 0x41, 0x40,
-        ])
-        .unwrap();
+        ];
 
-        let pubkey = secp256k1::PublicKey::from_secret_key(&s, &privkey);
+        let sk_ecdsa = secp256k1::SecretKey::from_slice(&KEY).unwrap();
+        let sk_schnorr =
+            secp256k1::schnorrsig::KeyPair::from_seckey_slice(&secp, &KEY)
+                .unwrap();
+
+        let pk_ecdsa = secp256k1::PublicKey::from_secret_key(&secp, &sk_ecdsa);
+        let pk_schnorr =
+            secp256k1::schnorrsig::PublicKey::from_keypair(&secp, &sk_schnorr);
         let msg = Message::from_slice(&[1u8; 32]).unwrap();
 
-        let sig = s.sign(&msg, &privkey);
-        let decoded_sig = test_suite(&sig, &SIG_BYTES, 64);
+        let ecdsa = secp.sign(&msg, &sk_ecdsa);
+        test_encoding_roundtrip(&ecdsa, &ECDSA_BYTES).unwrap();
+        assert!(secp.verify(&msg, &ecdsa, &pk_ecdsa).is_ok());
 
-        assert!(s.verify(&msg, &decoded_sig, &pubkey).is_ok());
+        let schnorr = secp.schnorrsig_sign_no_aux_rand(&msg, &sk_schnorr);
+        test_encoding_roundtrip(&schnorr, &SCHNORR_BYTES).unwrap();
+        assert!(secp.schnorrsig_verify(&schnorr, &msg, &pk_schnorr).is_ok());
+
+        // Schnorr signature can be deserialized as ECDSA and vice verse,
+        // (since there is no encoding-level way of verifying its type)
+        // but MUST be invalid upon signature validation
+        let schnorr_as_ecdsa: secp256k1::Signature =
+            test_vec_decoding_roundtrip(&SCHNORR_BYTES).unwrap();
+        let ecdsa_as_schnorr: secp256k1::schnorrsig::Signature =
+            test_vec_decoding_roundtrip(&ECDSA_BYTES).unwrap();
+        assert_eq!(
+            secp.verify(&msg, &schnorr_as_ecdsa, &pk_ecdsa),
+            Err(secp256k1::Error::IncorrectSignature)
+        );
+        assert_eq!(
+            secp.schnorrsig_verify(&ecdsa_as_schnorr, &msg, &pk_schnorr),
+            Err(secp256k1::Error::InvalidSignature)
+        );
     }
 
     #[test]
     #[should_panic(expected = "UnexpectedEof")]
-    fn test_garbagedata_sig() {
-        static SIG_BYTES: [u8; 58] = [
-            0xdf, 0x2b, 0x7, 0x1, 0x5f, 0x2e, 0x1, 0x67, 0x74, 0x18, 0x7e,
-            0xad, 0x4a, 0x4f, 0x71, 0x9a, 0x14, 0xe3, 0xe1, 0xad, 0xa1, 0x78,
-            0xd6, 0x6c, 0xce, 0xcf, 0xa4, 0x5b, 0x63, 0x30, 0x70, 0xc2, 0x43,
-            0xa2, 0xd7, 0x6e, 0xe0, 0x5d, 0x63, 0x49, 0xfe, 0x98, 0x69, 0x6c,
-            0x1c, 0x4d, 0x9a, 0x67, 0x11, 0x24, 0xde, 0x40, 0xc5, 0x31, 0x71,
-            0xa4, 0xb2, 0x82,
+    fn test_garbagedata_ecdsa() {
+        secp256k1::Signature::strict_decode(&ECDSA_BYTES[5..]).unwrap();
+    }
+
+    #[test]
+    #[should_panic(expected = "UnexpectedEof")]
+    fn test_garbagedata_schnorrsig() {
+        secp256k1::schnorrsig::Signature::strict_decode(&SCHNORR_BYTES[5..])
+            .unwrap();
+    }
+
+    #[test]
+    fn test_encoding_network(
+    ) -> Result<(), DataEncodingTestFailure<bitcoin::Network>> {
+        test_encoding_roundtrip(
+            &bitcoin::Network::Bitcoin,
+            &[0xF9, 0xBE, 0xB4, 0xD9],
+        )?;
+        test_encoding_roundtrip(
+            &bitcoin::Network::Testnet,
+            &[0x0B, 0x11, 0x09, 0x07],
+        )?;
+        test_encoding_roundtrip(
+            &bitcoin::Network::Signet,
+            &[0x0A, 0x03, 0xCF, 0x40],
+        )?;
+        test_encoding_roundtrip(
+            &bitcoin::Network::Regtest,
+            &[0xFA, 0xBF, 0xB5, 0xDA],
+        )
+    }
+
+    #[test]
+    #[should_panic(
+        expected = r#"ValueOutOfRange("bitcoin::Network", 0..0, 2762187425)"#
+    )]
+    fn test_encoding_network_failure() {
+        // Bitcoin Network structure do not support "Other" networks
+        bitcoin::Network::strict_decode(&[0xA1u8, 0xA2u8, 0xA3u8, 0xA4u8][..])
+            .unwrap();
+    }
+
+    #[test]
+    fn test_encoding_address() {
+        test_encoding_roundtrip(
+            &Address::from_str("12CL4K2eVqj7hQTix7dM7CVHCkpP17Pry3").unwrap(),
+            [
+                0xF9, 0xBE, 0xB4, 0xD9, 0x20, 0x0D, 0x1C, 0x9C, 0x02, 0xA7,
+                0xBE, 0x9B, 0xA8, 0xB8, 0x84, 0x28, 0x04, 0xFE, 0xB9, 0x61,
+                0x48, 0x1C, 0xE6, 0x56, 0x1B,
+            ],
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("3AfyxhpBVVLmBR4ZYX2onGzRqjv5QZ7FqD").unwrap(),
+            [
+                0xF9, 0xBE, 0xB4, 0xD9, 0x21, 0x62, 0x87, 0x13, 0xE2, 0x7A,
+                0x36, 0xDA, 0x16, 0x17, 0x4E, 0x9D, 0x02, 0xC1, 0x77, 0x2C,
+                0xD9, 0xE4, 0x06, 0x03, 0x9B,
+            ],
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("bc1qp5wfcq48h6d63wyy9qz0awtpfqwwv4sma86mhz")
+                .unwrap(),
+            [
+                0xF9, 0xBE, 0xB4, 0xD9, 0x00, 0x14, 0x00, 0x0D, 0x1C, 0x9C,
+                0x02, 0xA7, 0xBE, 0x9B, 0xA8, 0xB8, 0x84, 0x28, 0x04, 0xFE,
+                0xB9, 0x61, 0x48, 0x1C, 0xE6, 0x56, 0x1B,
+            ],
+        )
+        .unwrap();
+
+        static P2WSH_BC: [u8; 39] = [
+            0xF9, 0xBE, 0xB4, 0xD9, 0x00, 0x20, 0x00, 0x18, 0x63, 0x14, 0x3C,
+            0x14, 0xC5, 0x16, 0x68, 0x04, 0xBD, 0x19, 0x20, 0x33, 0x56, 0xDA,
+            0x13, 0x6C, 0x98, 0x56, 0x78, 0xCD, 0x4D, 0x27, 0xA1, 0xB8, 0xC6,
+            0x32, 0x96, 0x04, 0x90, 0x32, 0x62,
         ];
-        secp256k1::Signature::strict_decode(&SIG_BYTES[..]).unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("bc1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3qccfmv3")
+                .unwrap(),
+            P2WSH_BC,
+        ).unwrap();
+        // TODO: #18 test_encoding_roundtrip(&Address::from_str("bc1pw508d6qejxtdg4y5r3zarvary0c5xw7kw508d6qejxtdg4y5r3zarvary0c5xw7kt5nd6y").unwrap(), []).unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("mgiHMN7dJsANUWwLfgbiw7hc4kR5xMjPhw").unwrap(),
+            [
+                0x0B, 0x11, 0x09, 0x07, 0x20, 0x0D, 0x1C, 0x9C, 0x02, 0xA7,
+                0xBE, 0x9B, 0xA8, 0xB8, 0x84, 0x28, 0x04, 0xFE, 0xB9, 0x61,
+                0x48, 0x1C, 0xE6, 0x56, 0x1B,
+            ],
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("2N2EC2SkD6wr7PCh7DeegQDyh468FA7TK3a").unwrap(),
+            [
+                0x0B, 0x11, 0x09, 0x07, 0x21, 0x62, 0x87, 0x13, 0xE2, 0x7A,
+                0x36, 0xDA, 0x16, 0x17, 0x4E, 0x9D, 0x02, 0xC1, 0x77, 0x2C,
+                0xD9, 0xE4, 0x06, 0x03, 0x9B,
+            ],
+        )
+        .unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("tb1qp5wfcq48h6d63wyy9qz0awtpfqwwv4smhppgv3")
+                .unwrap(),
+            [
+                0x0B, 0x11, 0x09, 0x07, 0x00, 0x14, 0x00, 0x0D, 0x1C, 0x9C,
+                0x02, 0xA7, 0xBE, 0x9B, 0xA8, 0xB8, 0x84, 0x28, 0x04, 0xFE,
+                0xB9, 0x61, 0x48, 0x1C, 0xE6, 0x56, 0x1B,
+            ],
+        )
+        .unwrap();
+        static P2WSH_TB: [u8; 39] = [
+            0x0B, 0x11, 0x09, 0x07, 0x00, 0x20, 0x00, 0x18, 0x63, 0x14, 0x3C,
+            0x14, 0xC5, 0x16, 0x68, 0x04, 0xBD, 0x19, 0x20, 0x33, 0x56, 0xDA,
+            0x13, 0x6C, 0x98, 0x56, 0x78, 0xCD, 0x4D, 0x27, 0xA1, 0xB8, 0xC6,
+            0x32, 0x96, 0x04, 0x90, 0x32, 0x62,
+        ];
+        test_encoding_roundtrip(
+            &Address::from_str("tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7")
+                .unwrap(),
+            P2WSH_TB,
+        ).unwrap();
+        // TODO: #18 test_encoding_roundtrip(&Address::from_str("
+        // tb1pqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesf3hn0c").
+        // unwrap(), []).unwrap();
+        test_encoding_roundtrip(
+            &Address::from_str("bcrt1qs758ursh4q9z627kt3pp5yysm78ddny6txaqgw")
+                .unwrap(),
+            [
+                0xFA, 0xBF, 0xB5, 0xDA, 0x00, 0x14, 0x00, 0x87, 0xA8, 0x7E,
+                0x0E, 0x17, 0xA8, 0x0A, 0x2D, 0x2B, 0xD6, 0x5C, 0x42, 0x1A,
+                0x10, 0x90, 0xDF, 0x8E, 0xD6, 0xCC, 0x9A,
+            ],
+        )
+        .unwrap();
+    }
+
+    #[test]
+    #[should_panic(
+        expected = r#"ValueOutOfRange("witness program version", 0..17, 35)"#
+    )]
+    fn test_encoding_address_failure() {
+        // Address string with witness version byte (fifth) > 17 and not
+        // matching non-witness address type
+        Address::strict_deserialize([
+            0x0B, 0x11, 0x09, 0x07, 0x23, 0x14, 0x00, 0x0D, 0x1C, 0x9C, 0x02,
+            0xA7, 0xBE, 0x9B, 0xA8, 0xB8, 0x84, 0x28, 0x04, 0xFE, 0xB9, 0x61,
+            0x48, 0x1C, 0xE6, 0x56, 0x1B,
+        ])
+        .unwrap();
     }
 
     #[test]
@@ -599,9 +943,9 @@ pub(crate) mod test {
 
         // test random and null outpoints
         let outpoint = OutPoint::new(txid, vout);
-        let _ = test_suite(&outpoint, &OUTPOINT, 36);
+        let _ = test_encoding_roundtrip(&outpoint, &OUTPOINT).unwrap();
         let null = OutPoint::null();
-        let _ = test_suite(&null, &OUTPOINT_NULL, 36);
+        let _ = test_encoding_roundtrip(&null, &OUTPOINT_NULL).unwrap();
     }
 
     #[test]
@@ -615,6 +959,15 @@ pub(crate) mod test {
         OutPoint::strict_decode(&OUTPOINT[..]).unwrap();
     }
 
+    #[test]
+    fn test_amount() {
+        let value = 19_356_465_2435_5767__u64;
+        let amount = Amount::from_sat(value);
+        let data = value.to_le_bytes();
+        test_encoding_roundtrip(&value, data).unwrap();
+        test_encoding_roundtrip(&amount, data).unwrap();
+    }
+
     #[test]
     fn test_tx() {
         let tx_segwit_bytes = Vec::from_hex(
@@ -645,12 +998,9 @@ pub(crate) mod test {
         let tx_legacy2: Transaction =
             consensus::deserialize(&tx_legacy2_bytes).unwrap();
 
-        assert_eq!(strict_serialize(&tx_segwit).unwrap(), tx_segwit_bytes);
-        assert_eq!(strict_serialize(&tx_legacy1).unwrap(), tx_legacy1_bytes);
-        assert_eq!(strict_serialize(&tx_legacy2).unwrap(), tx_legacy2_bytes);
-        test_suite(&tx_segwit, &tx_segwit_bytes, tx_segwit_bytes.len());
-        test_suite(&tx_legacy1, &tx_legacy1_bytes, tx_legacy1_bytes.len());
-        test_suite(&tx_legacy2, &tx_legacy2_bytes, tx_legacy2_bytes.len());
+        test_encoding_roundtrip(&tx_segwit, &tx_segwit_bytes).unwrap();
+        test_encoding_roundtrip(&tx_legacy1, &tx_legacy1_bytes).unwrap();
+        test_encoding_roundtrip(&tx_legacy2, &tx_legacy2_bytes).unwrap();
     }
 
     #[test]
@@ -663,8 +1013,7 @@ pub(crate) mod test {
             0711c06c7f3e097c9447c52ffffffff"
         ).unwrap();
         let txin: TxIn = consensus::deserialize(&txin_bytes).unwrap();
-        assert_eq!(strict_serialize(&txin).unwrap(), txin_bytes);
-        test_suite(&txin, &txin_bytes, txin_bytes.len());
+        test_encoding_roundtrip(&txin, &txin_bytes).unwrap();
     }
 
     #[test]
@@ -683,24 +1032,8 @@ pub(crate) mod test {
         let txout_legacy: TxOut =
             consensus::deserialize(&txout_legacy_bytes).unwrap();
 
-        assert_eq!(
-            strict_serialize(&txout_segwit).unwrap(),
-            txout_segwit_bytes
-        );
-        assert_eq!(
-            strict_serialize(&txout_legacy).unwrap(),
-            txout_legacy_bytes
-        );
-        test_suite(
-            &txout_segwit,
-            &txout_segwit_bytes,
-            txout_segwit_bytes.len(),
-        );
-        test_suite(
-            &txout_legacy,
-            &txout_legacy_bytes,
-            txout_legacy_bytes.len(),
-        );
+        test_encoding_roundtrip(&txout_segwit, &txout_segwit_bytes).unwrap();
+        test_encoding_roundtrip(&txout_legacy, &txout_legacy_bytes).unwrap();
     }
 
     #[test]
@@ -728,8 +1061,7 @@ pub(crate) mod test {
         let psbt: PartiallySignedTransaction =
             consensus::deserialize(&psbt_bytes).unwrap();
 
-        assert_eq!(strict_serialize(&psbt).unwrap(), psbt_bytes);
-        test_suite(&psbt, &psbt_bytes, psbt_bytes.len());
+        test_encoding_roundtrip(&psbt, &psbt_bytes).unwrap();
     }
 
     #[test]
@@ -756,14 +1088,14 @@ pub(crate) mod test {
             9ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8",
         )
         .unwrap();
-        test_suite(&ext_pubkey1, &EXT_PUBKEY1, 78);
+        test_encoding_roundtrip(&ext_pubkey1, &EXT_PUBKEY1).unwrap();
 
         let ext_pubkey2 = bip32::ExtendedPubKey::from_str(
             "xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJP\
             MM3No2dFDFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5",
         )
         .unwrap();
-        test_suite(&ext_pubkey2, &EXT_PUBKEY2, 78);
+        test_encoding_roundtrip(&ext_pubkey2, &EXT_PUBKEY2).unwrap();
     }
 
     #[test]
@@ -804,33 +1136,28 @@ pub(crate) mod test {
         ];
 
         // OP_RETURN
-        let op_return = Script::strict_decode(&OP_RETURN[..]).unwrap();
-        let decode_opreturn = test_suite(&op_return, &OP_RETURN, 40);
-        assert!(decode_opreturn.is_op_return());
+        let op_return: Script =
+            test_vec_decoding_roundtrip(&OP_RETURN).unwrap();
+        assert!(op_return.is_op_return());
 
         // P2PK
-        let p2pk = Script::strict_decode(&P2PK[..]).unwrap();
-        let decode_p2pk = test_suite(&p2pk, &P2PK, 37);
-        assert!(decode_p2pk.is_p2pk());
+        let p2pk: Script = test_vec_decoding_roundtrip(&P2PK).unwrap();
+        assert!(p2pk.is_p2pk());
 
         //P2PKH
-        let p2pkh = Script::strict_decode(&P2PKH[..]).unwrap();
-        let decode_p2pkh = test_suite(&p2pkh, &P2PKH, 27);
-        assert!(decode_p2pkh.is_p2pkh());
+        let p2pkh: Script = test_vec_decoding_roundtrip(&P2PKH).unwrap();
+        assert!(p2pkh.is_p2pkh());
 
         //P2SH
-        let p2sh = Script::strict_decode(&P2SH[..]).unwrap();
-        let decode_p2sh = test_suite(&p2sh, &P2SH, 25);
-        assert!(decode_p2sh.is_p2sh());
+        let p2sh: Script = test_vec_decoding_roundtrip(&P2SH).unwrap();
+        assert!(p2sh.is_p2sh());
 
         //P2WPKH
-        let p2wpkh = Script::strict_decode(&P2WPKH[..]).unwrap();
-        let decode_p2wpkh = test_suite(&p2wpkh, &P2WPKH, 24);
-        assert!(decode_p2wpkh.is_v0_p2wpkh());
+        let p2wpkh: Script = test_vec_decoding_roundtrip(&P2WPKH).unwrap();
+        assert!(p2wpkh.is_v0_p2wpkh());
 
         //P2WSH
-        let p2wsh = Script::strict_decode(&P2WSH[..]).unwrap();
-        let decoded_p2wsh = test_suite(&p2wsh, &P2WSH, 36);
-        assert!(decoded_p2wsh.is_v0_p2wsh());
+        let p2wsh: Script = test_vec_decoding_roundtrip(&P2WSH).unwrap();
+        assert!(p2wsh.is_v0_p2wsh());
     }
 }
