@@ -1,5 +1,5 @@
-// LNP/BP client-side-validation library implementing respective LNPBP
-// specifications & standards (LNPBP-7, 8, 9, 42)
+// LNP/BP client-side-validation foundation libraries implementing LNPBP
+// specifications & standards (LNPBP-4, 7, 8, 9, 42, 81)
 //
 // Written in 2019-2021 by
 //     Dr. Maxim Orlovsky <orlovsky@pandoracore.com>
@@ -737,9 +737,7 @@ impl Strategy for SocketAddrV6 {
 mod test {
     use super::*;
     use crate::strict_deserialize;
-    use bitcoin::secp256k1::PublicKey;
     use std::convert::TryInto;
-    use std::str::FromStr;
 
     fn gen_ipv4_addrs() -> Vec<Ipv4Addr> {
         let vars = [0u8, 1, 32, 48, 64, 127, 168, 192, 254, 255];
@@ -808,8 +806,15 @@ mod test {
         assert_eq!(OnionV3 as u8, 3);
         assert_eq!(Lightning as u8, 4);
 
-        test_enum_u8_exhaustive!(crate => AddrFormat; IpV4 => 0, IpV6 => 1, OnionV2 => 2, OnionV3 => 3, Lightning => 4);
-        test_enum_u8_exhaustive!(crate => Transport; Tcp => 1, Udp => 2, Mtcp => 3, Quic => 4);
+        test_encoding_enum_u8_exhaustive!(
+            crate => AddrFormat;
+            IpV4 => 0u8, IpV6 => 1u8, OnionV2 => 2u8, OnionV3 => 3u8, Lightning => 4u8
+        ).unwrap();
+        test_encoding_enum_u8_exhaustive!(
+            crate => Transport;
+            Tcp => 1u8, Udp => 2u8, Mtcp => 3u8, Quic => 4u8
+        )
+        .unwrap();
     }
 
     #[test]
@@ -1099,7 +1104,11 @@ mod test {
     }
 
     #[test]
+    #[cfg(feature = "bitcoin")]
     fn uniform_raw_roundtrip_other() {
+        use bitcoin::secp256k1::PublicKey;
+        use std::str::FromStr;
+
         let lk = PublicKey::from_str("02d1780dd0e08f4d873f94faf49d878d909a1174291d3fcac3e02a6c45e7eda744").unwrap();
         let addr = lk.serialize();
 
@@ -1139,7 +1148,7 @@ mod test {
 
     #[test]
     fn uniform_raw_conversion() {
-        use bitcoin::hashes::hex::ToHex;
+        use bitcoin_hashes::hex::ToHex;
         use std::str::FromStr;
 
         let ips = vec![
